# coding: utf-8
""" Обновляет устаревшие записи в кэше DNS """

import os
import json
import shutil
import logging
from random import randint
from tempfile import mkstemp
from datetime import datetime
import qdns
from dns_config import db_file, list_file


__author__ = 'oleg'


CURRENT_UNIXTIME = lambda: int(datetime.now().strftime("%s"))
MIN_TTL = 5
MAX_TTL = int(os.environ.get('MAX_TTL', 60))
NXDOMAIN_TTL = int(os.environ.get('NXDOMAIN_TTL', 180))
IP_TTL = int(os.environ.get('IP_TTL', 720))
DOMAINS = set([unicode(line, 'utf-8').strip() for line in open(list_file).readlines()])


log = logging.getLogger('qdns')
log_level = logging.INFO
if os.environ.get('RESOLVER_DEBUG') == '1':
    log_level = logging.DEBUG
logging.basicConfig(format='%(asctime)-15s %(message)s', level=log_level)

DNS_BLACKHOLES = ['127.0.0.1', '95.213.143.204']
if os.environ.get('DNS_BLACKHOLE_IP'):
    DNS_BLACKHOLES.append(os.environ.get('DNS_BLACKHOLE_IP'))


QUEUED_COUNT = 0
QUEUED_COUNT_TOTAL = 0

def eval_ttl(dirty, ttl):
    """растёт если ничего не меняется, уменьшается если часто меняется, MIN_TTL<ttl<MAX_TTL"""
    if ttl == NXDOMAIN_TTL:
        return NXDOMAIN_TTL
    if dirty:
        ttl -= 10
    else:
        ttl += randint(1, 5)
    if ttl > MAX_TTL:
        ttl = MAX_TTL
    if ttl < MIN_TTL:
        ttl = MIN_TTL
    return ttl


def process_ip_lists(entry, ip_list, ip6_list):
    """заполняет списки для записи и возвращает флаг изменились ли они или нет"""
    last_update = CURRENT_UNIXTIME()
    dirty = False
    log.debug("input: ipv4: %s, ipv6: %s", ip_list, ip6_list)
    for version in ('ipv4', 'ipv6'):
        log.debug("process %s", version)
        if version == 'ipv4':
            received = [ipv4 for ipv4 in ip_list if ipv4 not in DNS_BLACKHOLES]
            out = entry.get('ip', {})
        elif version == 'ipv6':
            received = ip6_list
            out = entry.get('ip6', {})
        log.debug("in: %s, out: %s", received, out)
        out_len = len(out)
        for ip in received:
            out[ip] = (IP_TTL, last_update)
        if out_len != len(out):
            logging.debug('dirty cause changed len')
            dirty = True
        for ip, val in out.items():
            if val[1] + val[0] * 60 < last_update:
                logging.info('delete %s', ip)
                del out[ip]
                dirty = True
                logging.debug('dirty cause del item')
    return dirty, last_update


def update_entry(domain, ip, ip6, ttl=None):
    """ Вставляем в dict новый dict о домене """
    global QUEUED_COUNT
    global QUEUED_COUNT_TOTAL
    domain = domain.lower()
    entry = DB.get(domain)
    if not entry:
        return

    ttl = ttl or entry.get('ttl', 5)
    ip.sort()
    ip6.sort()
    dirty, last_update = process_ip_lists(entry, ip, ip6)
    ttl = eval_ttl(dirty, ttl)
    if dirty:
        logging.info("updated: %s (%d/%d remains)", domain.encode('utf-8'), QUEUED_COUNT, QUEUED_COUNT_TOTAL)
    else:
        logging.debug("not changed: %s (%d/%d remains)", domain.encode('utf-8'), QUEUED_COUNT, QUEUED_COUNT_TOTAL)

    DB[domain] = {
        "ip": entry.get('ip', {}),
        "ip6": entry.get('ip6', {}),
        "last_update": last_update,
        "ttl": ttl,
    }
    log.debug("%s: %s", domain, DB[domain])


def receive(result):
    """ Callback для QDNS, возвращает name с запросом и ответом """
    global QUEUED_COUNT
    QUEUED_COUNT -= 1
    if result:
        name, ip_list, ip6_list = result
        update_entry(name, ip_list, ip6_list)


def is_out_date(entry, cur_time):
    """ Проверяем, устарели ли данные о домене и насколько """
    if not (entry.get('ttl') and entry.get('last_update')):
        return True
    if entry.get('ttl') * 60 + entry.get('last_update') < cur_time:
        return True
    out_of_date = any(
        [ip[1] + ip[0] * 60 < cur_time for ip in entry.get('ip').values()])
    return out_of_date


def queue_normal():
    """ Прогоняем все домены резолвером """
    global QUEUED_COUNT
    global QUEUED_COUNT_TOTAL
    qdns_configured = False
    current_time = CURRENT_UNIXTIME()
    logging.info('queue outdated domains')
    QUEUED_COUNT = 0
    for domain, entry in DB.items():
        if domain not in DOMAINS:
            continue
        if not is_out_date(entry, current_time):
            continue
        if not qdns_configured:
            logging.info('qdns.configure')
            cpu_count = int(os.environ.get('QDNS_THREADS', 4))
            qdns.configure(new_thread_count=cpu_count, carbon=True)
            qdns_configured = True
        QUEUED_COUNT += 1
        logging.info("queued: %s", domain.encode('utf-8'))
        qdns.gethostsbyname(domain, receive)
    QUEUED_COUNT_TOTAL = int(QUEUED_COUNT)
    logging.info("total queued: %d", QUEUED_COUNT_TOTAL)
    if qdns_configured:
        qdns.stop(True)
        logging.info('qdns stopped')


def __set_as_bad(domain, entry):
    logging.info("unresolved %s ttl %d->%d", domain, entry.get('ttl'), NXDOMAIN_TTL)
    update_entry(domain, [], [], NXDOMAIN_TTL)


def set_as_bad():
    """ Помечаем домены, по которым не получили ответ от DNS сервера плохими """
    logging.info('mark unresolved domains')
    current_time = CURRENT_UNIXTIME()
    for domain, entry in DB.items():
        if domain not in DOMAINS:
            continue
        out_of_date = is_out_date(entry, current_time)
        empty = not entry.get('ip') and not entry.get('ip6')
        marked_as_bad = entry.get('ttl') == NXDOMAIN_TTL
        if out_of_date or (empty and not marked_as_bad):
            __set_as_bad(domain, entry)


def remove_blackhole_ip():
    """remove blackholes ip addresses from resolv cache"""
    logging.info("%s", remove_blackhole_ip.__doc__)
    for domain, value in DB.items():
        for blackhole in DNS_BLACKHOLES:
            if value.get('ip').get(blackhole):
                logging.info('remove %s from %s', blackhole, domain)
                del value.get('ip')[blackhole]


def save(db_data):
    """ сохраняем отрезолвленное в json-file """
    _, fpath = mkstemp(prefix='/tmp/https_resolv.json.')
    with open(fpath, 'w') as tmpfile:
        json.dump(db_data, tmpfile, indent=4)
    shutil.copyfile(fpath, db_file)  # json.dump + диагностика

if __name__ == '__main__':
    DB = json.load(open(db_file))
    logging.info('start')
    queue_normal()
    set_as_bad()
    remove_blackhole_ip()
    save(DB)
    logging.info('finished')
